@@ -1,30 +1,50 @@
-import ctypes, ctypes.util
+import ctypes, ctypes.util, operator, sys
 from . import model
 
+if sys.version_info < (3,):
+    integer_types = (int, long)
+    bytechr = chr
+else:
+    unicode = str
+    integer_types = int
+    xrange = range
+    bytechr = lambda num: bytes([num])
+
 class CTypesData(object):
-    __slots__ = []
+    __slots__ = ['__weakref__']
 
     def __init__(self, *args):
         raise TypeError("cannot instantiate %r" % (self.__class__,))
 
+    @classmethod
+    def _newp(cls, init):
+        raise TypeError("expected a pointer or array ctype, got '%s'"
+                        % (cls._get_c_name(),))
+
     @staticmethod
     def _to_ctypes(value):
         raise TypeError
 
     @classmethod
-    def _arg_to_ctypes(cls, value):
+    def _arg_to_ctypes(cls, *value):
         try:
             ctype = cls._ctype
         except AttributeError:
             raise TypeError("cannot create an instance of %r" % (cls,))
-        res = cls._to_ctypes(value)
-        if not isinstance(res, ctype):
-            res = cls._ctype(res)
+        if value:
+            res = cls._to_ctypes(*value)
+            if not isinstance(res, ctype):
+                res = cls._ctype(res)
+        else:
+            res = cls._ctype()
         return res
 
     @classmethod
     def _create_ctype_obj(cls, init):
-        return cls._arg_to_ctypes(init)
+        if init is None:
+            return cls._arg_to_ctypes()
+        else:
+            return cls._arg_to_ctypes(init)
 
     @staticmethod
     def _from_ctypes(ctypes_value):
@@ -37,14 +57,23 @@ class CTypesData(object):
     @classmethod
     def _fix_class(cls):
         cls.__name__ = 'CData<%s>' % (cls._get_c_name(),)
+        cls.__qualname__ = 'CData<%s>' % (cls._get_c_name(),)
         cls.__module__ = 'ffi'
 
     def _get_own_repr(self):
-        return ''
+        raise NotImplementedError
+
+    def _addr_repr(self, address):
+        if address == 0:
+            return 'NULL'
+        else:
+            if address < 0:
+                address += 1 << (8*ctypes.sizeof(ctypes.c_void_p))
+            return '0x%x' % address
 
     def __repr__(self, c_name=None):
         own = self._get_own_repr()
-        return '<cdata %r%s>' % (c_name or self._get_c_name(), own)
+        return '<cdata %r %s>' % (c_name or self._get_c_name(), own)
 
     def _convert_to_address(self, BClass):
         if BClass is None:
@@ -76,30 +105,77 @@ class CTypesData(object):
         raise TypeError("cdata %r does not support iteration" % (
             self._get_c_name()),)
 
+    def _make_cmp(name):
+        cmpfunc = getattr(operator, name)
+        def cmp(self, other):
+            if isinstance(other, CTypesData):
+                return cmpfunc(self._convert_to_address(None),
+                               other._convert_to_address(None))
+            else:
+                return NotImplemented
+        cmp.func_name = name
+        return cmp
+
+    __eq__ = _make_cmp('__eq__')
+    __ne__ = _make_cmp('__ne__')
+    __lt__ = _make_cmp('__lt__')
+    __le__ = _make_cmp('__le__')
+    __gt__ = _make_cmp('__gt__')
+    __ge__ = _make_cmp('__ge__')
+
+    def __hash__(self):
+        return hash(type(self)) ^ hash(self._convert_to_address(None))
+
+    def _to_string(self, maxlen):
+        raise TypeError("string(): %r" % (self,))
+
 
 class CTypesGenericPrimitive(CTypesData):
     __slots__ = []
 
+    def __eq__(self, other):
+        return self is other
+
+    def __ne__(self, other):
+        return self is not other
+
+    def __hash__(self):
+        return object.__hash__(self)
+
+    def _get_own_repr(self):
+        return repr(self._from_ctypes(self._value))
+
 
 class CTypesGenericArray(CTypesData):
     __slots__ = []
 
+    @classmethod
+    def _newp(cls, init):
+        return cls(init)
+
     def __iter__(self):
         for i in xrange(len(self)):
             yield self[i]
 
+    def _get_own_repr(self):
+        return self._addr_repr(ctypes.addressof(self._blob))
+
 
 class CTypesGenericPtr(CTypesData):
     __slots__ = ['_address', '_as_ctype_ptr']
     _automatic_casts = False
 
+    @classmethod
+    def _newp(cls, init):
+        return cls(init)
+
     @classmethod
     def _cast_from(cls, source):
         if source is None:
             address = 0
         elif isinstance(source, CTypesData):
             address = source._cast_to_integer()
-        elif isinstance(source, (int, long)):
+        elif isinstance(source, integer_types):
             address = source
         else:
             raise TypeError("bad type for cast to %r: %r" %
@@ -113,46 +189,41 @@ class CTypesGenericPtr(CTypesData):
         self._as_ctype_ptr = ctypes.cast(address, cls._ctype)
         return self
 
+    def _get_own_repr(self):
+        try:
+            return self._addr_repr(self._address)
+        except AttributeError:
+            return '???'
+
     def _cast_to_integer(self):
         return self._address
 
     def __nonzero__(self):
         return bool(self._address)
-
-    def __eq__(self, other):
-        if other is None:
-            return not bool(self._address)
-        return (type(self) is type(other) and
-                self._address == other._address)
-
-    def __ne__(self, other):
-        return not self.__eq__(other)
-
-    def __hash__(self):
-        return hash(type(self)) ^ hash(self._address)
+    
+    def __bool__(self):
+        return bool(self._address)
 
     @classmethod
     def _to_ctypes(cls, value):
-        if value is None:
-            address = 0
-        else:
-            address = value._convert_to_address(cls)
+        if not isinstance(value, CTypesData):
+            raise TypeError("unexpected %s object" % type(value).__name__)
+        address = value._convert_to_address(cls)
         return ctypes.cast(address, cls._ctype)
 
     @classmethod
     def _from_ctypes(cls, ctypes_ptr):
-        if not ctypes_ptr:
-            return None
         address = ctypes.cast(ctypes_ptr, ctypes.c_void_p).value or 0
         return cls._new_pointer_at(address)
 
     @classmethod
     def _initialize(cls, ctypes_ptr, value):
-        if value is not None:
+        if value:
             ctypes_ptr.contents = cls._to_ctypes(value).contents
 
     def _convert_to_address(self, BClass):
-        if BClass in (self.__class__, None) or BClass._automatic_casts:
+        if (BClass in (self.__class__, None) or BClass._automatic_casts
+            or self._automatic_casts):
             return self._address
         else:
             return CTypesData._convert_to_address(self, BClass)
@@ -166,6 +237,9 @@ class CTypesBaseStructOrUnion(CTypesData):
         # may be overridden
         raise TypeError("cannot instantiate opaque type %s" % (cls,))
 
+    def _get_own_repr(self):
+        return self._addr_repr(ctypes.addressof(self._blob))
+
     @classmethod
     def _offsetof(cls, fieldname):
         return getattr(cls._ctype, fieldname).offset
@@ -186,6 +260,9 @@ class CTypesBaseStructOrUnion(CTypesData):
     def _to_ctypes(cls, value):
         return value._blob
 
+    def __repr__(self, c_name=None):
+        return CTypesData.__repr__(self, c_name or self._get_c_name(' &'))
+
 
 class CTypesBackend(object):
 
@@ -222,8 +299,6 @@ class CTypesBackend(object):
             if size == ctypes.sizeof(ctypes.c_size_t):
                 result['size_t'] = size | UNSIGNED
                 result['ssize_t'] = size
-            #if size == ctypes.sizeof(ctypes.c_wchar):
-            #    result['wchar_t'] = size | UNSIGNED
         return result
 
     def load_library(self, path):
@@ -239,26 +314,34 @@ class CTypesBackend(object):
                 return None
             @staticmethod
             def _to_ctypes(novalue):
+                if novalue is not None:
+                    raise TypeError("None expected, got %s object" %
+                                    (type(novalue).__name__,))
                 return None
         CTypesVoid._fix_class()
         return CTypesVoid
 
     def new_primitive_type(self, name):
+        if name == 'wchar_t':
+            raise NotImplementedError(name)
         ctype = self.PRIMITIVE_TYPES[name]
         if name == 'char':
             kind = 'char'
         elif name in ('float', 'double'):
             kind = 'float'
         else:
-            kind = 'int'
+            if name in ('signed char', 'unsigned char'):
+                kind = 'byte'
+            else:
+                kind = 'int'
             is_signed = (ctype(-1).value == -1)
         #
         def _cast_source_to_int(source):
-            if isinstance(source, (int, long, float)):
+            if isinstance(source, (integer_types, float)):
                 source = int(source)
             elif isinstance(source, CTypesData):
                 source = source._cast_to_integer()
-            elif isinstance(source, str):
+            elif isinstance(source, bytes):
                 source = ord(source)
             elif source is None:
                 source = 0
@@ -281,7 +364,7 @@ class CTypesBackend(object):
                     return ctype()
                 return ctype(CTypesPrimitive._to_ctypes(init))
 
-            if kind == 'int':
+            if kind == 'int' or kind == 'byte':
                 @classmethod
                 def _cast_from(cls, source):
                     source = _cast_source_to_int(source)
@@ -294,12 +377,10 @@ class CTypesBackend(object):
                 @classmethod
                 def _cast_from(cls, source):
                     source = _cast_source_to_int(source)
-                    source = chr(source & 0xFF)
+                    source = bytechr(source & 0xFF)
                     return cls(source)
                 def __int__(self):
                     return ord(self._value)
-                def __str__(self):
-                    return self._value
 
             if kind == 'float':
                 @classmethod
@@ -322,10 +403,10 @@ class CTypesBackend(object):
 
             _cast_to_integer = __int__
 
-            if kind == 'int':
+            if kind == 'int' or kind == 'byte':
                 @staticmethod
                 def _to_ctypes(x):
-                    if not isinstance(x, (int, long)):
+                    if not isinstance(x, integer_types):
                         if isinstance(x, CTypesData):
                             x = int(x)
                         else:
@@ -342,7 +423,7 @@ class CTypesBackend(object):
             if kind == 'char':
                 @staticmethod
                 def _to_ctypes(x):
-                    if isinstance(x, str) and len(x) == 1:
+                    if isinstance(x, bytes) and len(x) == 1:
                         return x
                     if isinstance(x, CTypesPrimitive):    # <CData <char>>
                         return x._value
@@ -352,7 +433,7 @@ class CTypesBackend(object):
             if kind == 'float':
                 @staticmethod
                 def _to_ctypes(x):
-                    if not isinstance(x, (int, long, float, CTypesData)):
+                    if not isinstance(x, (integer_types, float, CTypesData)):
                         raise TypeError("float expected, got %s" %
                                         type(x).__name__)
                     return ctype(x).value
@@ -364,18 +445,31 @@ class CTypesBackend(object):
             @staticmethod
             def _initialize(blob, init):
                 blob.value = CTypesPrimitive._to_ctypes(init)
+
+            if kind == 'char':
+                def _to_string(self, maxlen):
+                    return self._value
+            if kind == 'byte':
+                def _to_string(self, maxlen):
+                    return chr(self._value & 0xff)
         #
         CTypesPrimitive._fix_class()
         return CTypesPrimitive
 
     def new_pointer_type(self, BItem):
-        if BItem is self.ffi._get_cached_btype(model.PrimitiveType('char')):
+        getbtype = self.ffi._get_cached_btype
+        if BItem is getbtype(model.PrimitiveType('char')):
             kind = 'charp'
+        elif BItem in (getbtype(model.PrimitiveType('signed char')),
+                       getbtype(model.PrimitiveType('unsigned char'))):
+            kind = 'bytep'
         else:
             kind = 'generic'
         #
         class CTypesPtr(CTypesGenericPtr):
             __slots__ = ['_own']
+            if kind == 'charp':
+                __slots__ += ['__as_strbuf']
             _BItem = BItem
             if hasattr(BItem, '_ctype'):
                 _ctype = ctypes.POINTER(BItem._ctype)
@@ -389,20 +483,26 @@ class CTypesBackend(object):
 
             def __init__(self, init):
                 ctypeobj = BItem._create_ctype_obj(init)
-                self._as_ctype_ptr = ctypes.pointer(ctypeobj)
+                if kind == 'charp':
+                    self.__as_strbuf = ctypes.create_string_buffer(
+                        ctypeobj.value + b'\x00')
+                    self._as_ctype_ptr = ctypes.cast(
+                        self.__as_strbuf, self._ctype)
+                else:
+                    self._as_ctype_ptr = ctypes.pointer(ctypeobj)
                 self._address = ctypes.cast(self._as_ctype_ptr,
                                             ctypes.c_void_p).value
                 self._own = True
 
             def __add__(self, other):
-                if isinstance(other, (int, long)):
+                if isinstance(other, integer_types):
                     return self._new_pointer_at(self._address +
                                                 other * self._bitem_size)
                 else:
                     return NotImplemented
 
             def __sub__(self, other):
-                if isinstance(other, (int, long)):
+                if isinstance(other, integer_types):
                     return self._new_pointer_at(self._address -
                                                 other * self._bitem_size)
                 elif type(self) is type(other):
@@ -419,23 +519,29 @@ class CTypesBackend(object):
                 self._as_ctype_ptr[index] = BItem._to_ctypes(value)
 
             if kind == 'charp':
-                def __str__(self):
-                    n = 0
-                    while self._as_ctype_ptr[n] != '\x00':
-                        n += 1
-                    return ''.join([self._as_ctype_ptr[i] for i in range(n)])
                 @classmethod
                 def _arg_to_ctypes(cls, value):
-                    if isinstance(value, str):
+                    if isinstance(value, bytes):
                         return ctypes.c_char_p(value)
                     else:
                         return super(CTypesPtr, cls)._arg_to_ctypes(value)
 
+            if kind == 'charp' or kind == 'bytep':
+                def _to_string(self, maxlen):
+                    if maxlen < 0:
+                        maxlen = sys.maxsize
+                    p = ctypes.cast(self._as_ctype_ptr,
+                                    ctypes.POINTER(ctypes.c_char))
+                    n = 0
+                    while n < maxlen and p[n] != b'\x00':
+                        n += 1
+                    return b''.join([p[i] for i in range(n)])
+
             def _get_own_repr(self):
                 if getattr(self, '_own', False):
-                    return ' owning %d bytes' % (
+                    return 'owning %d bytes' % (
                         ctypes.sizeof(self._as_ctype_ptr.contents),)
-                return ''
+                return super(CTypesPtr, self)._get_own_repr()
         #
         if (BItem is self.ffi._get_cached_btype(model.void_type) or
             BItem is self.ffi._get_cached_btype(model.PrimitiveType('char'))):
@@ -450,8 +556,12 @@ class CTypesBackend(object):
         else:
             brackets = ' &[%d]' % length
         BItem = CTypesPtr._BItem
-        if BItem is self.ffi._get_cached_btype(model.PrimitiveType('char')):
+        getbtype = self.ffi._get_cached_btype
+        if BItem is getbtype(model.PrimitiveType('char')):
             kind = 'char'
+        elif BItem in (getbtype(model.PrimitiveType('signed char')),
+                       getbtype(model.PrimitiveType('unsigned char'))):
+            kind = 'byte'
         else:
             kind = 'generic'
         #
@@ -462,16 +572,19 @@ class CTypesBackend(object):
             else:
                 __slots__.append('_ctype')
             _reftypename = BItem._get_c_name(brackets)
+            _declared_length = length
+            _CTPtr = CTypesPtr
 
             def __init__(self, init):
                 if length is None:
-                    if isinstance(init, (int, long)):
+                    if isinstance(init, integer_types):
                         len1 = init
                         init = None
+                    elif kind == 'char' and isinstance(init, bytes):
+                        len1 = len(init) + 1    # extra null
                     else:
-                        extra_null = (kind == 'char' and isinstance(init, str))
                         init = tuple(init)
-                        len1 = len(init) + extra_null
+                        len1 = len(init)
                     self._ctype = BItem._ctype * len1
                 self._blob = self._ctype()
                 self._own = True
@@ -480,7 +593,10 @@ class CTypesBackend(object):
 
             @staticmethod
             def _initialize(blob, init):
-                init = tuple(init)
+                if isinstance(init, bytes):
+                    init = [init[i:i+1] for i in range(len(init))]
+                else:
+                    init = tuple(init)
                 if len(init) > len(blob):
                     raise IndexError("too many initializers")
                 addr = ctypes.cast(blob, ctypes.c_void_p).value
@@ -503,19 +619,21 @@ class CTypesBackend(object):
                     raise IndexError
                 self._blob[index] = BItem._to_ctypes(value)
 
-            if kind == 'char':
-                def __str__(self):
-                    s = ''.join(self._blob)
-                    try:
-                        s = s[:s.index('\x00')]
-                    except ValueError:
-                        pass
-                    return s
+            if kind == 'char' or kind == 'byte':
+                def _to_string(self, maxlen):
+                    if maxlen < 0:
+                        maxlen = len(self._blob)
+                    p = ctypes.cast(self._blob,
+                                    ctypes.POINTER(ctypes.c_char))
+                    n = 0
+                    while n < maxlen and p[n] != b'\x00':
+                        n += 1
+                    return b''.join([p[i] for i in range(n)])
 
             def _get_own_repr(self):
                 if getattr(self, '_own', False):
-                    return ' owning %d bytes' % (ctypes.sizeof(self._blob),)
-                return ''
+                    return 'owning %d bytes' % (ctypes.sizeof(self._blob),)
+                return super(CTypesArray, self)._get_own_repr()
 
             def _convert_to_address(self, BClass):
                 if BClass in (CTypesPtr, None) or BClass._automatic_casts:
@@ -534,12 +652,17 @@ class CTypesBackend(object):
                 return CTypesPtr._arg_to_ctypes(value)
 
             def __add__(self, other):
-                if isinstance(other, (int, long)):
+                if isinstance(other, integer_types):
                     return CTypesPtr._new_pointer_at(
                         ctypes.addressof(self._blob) +
                         other * ctypes.sizeof(BItem._ctype))
                 else:
                     return NotImplemented
+
+            @classmethod
+            def _cast_from(cls, source):
+                raise NotImplementedError("casting to %r" % (
+                    cls._get_c_name(),))
         #
         CTypesArray._fix_class()
         return CTypesArray
@@ -565,7 +688,12 @@ class CTypesBackend(object):
     def new_union_type(self, name):
         return self._new_struct_or_union('union', name, ctypes.Union)
 
-    def complete_struct_or_union(self, CTypesStructOrUnion, fields, tp):
+    def complete_struct_or_union(self, CTypesStructOrUnion, fields, tp,
+                                 totalsize=-1, totalalignment=-1):
+        if totalsize >= 0 or totalalignment >= 0:
+            raise NotImplementedError("the ctypes backend of CFFI does not support "
+                                      "structures completed by verify(); please "
+                                      "compile and install the _cffi_backend module.")
         struct_or_union = CTypesStructOrUnion._ctype
         fnames = [fname for (fname, BField, bitsize) in fields]
         btypes = [BField for (fname, BField, bitsize) in fields]
@@ -587,32 +715,31 @@ class CTypesBackend(object):
             return result
         CTypesStructOrUnion._create_ctype_obj = _create_ctype_obj
         #
-        if CTypesStructOrUnion._kind == 'struct':
-            def initialize(blob, init):
-                if not isinstance(init, dict):
-                    init = tuple(init)
-                    if len(init) > len(fnames):
-                        raise ValueError("too many values for %s initializer" %
-                                         CTypesStructOrUnion._get_c_name())
-                    init = dict(zip(fnames, init))
-                addr = ctypes.addressof(blob)
-                for fname, value in init.items():
-                    BField, bitsize = name2fieldtype[fname]
-                    assert bitsize < 0, \
-                           "not implemented: initializer with bit fields"
-                    offset = CTypesStructOrUnion._offsetof(fname)
-                    PTR = ctypes.POINTER(BField._ctype)
-                    p = ctypes.cast(addr + offset, PTR)
-                    BField._initialize(p.contents, value)
-            name2fieldtype = dict(zip(fnames, zip(btypes, bitfields)))
-        #
-        if CTypesStructOrUnion._kind == 'union':
-            def initialize(blob, init):
-                addr = ctypes.addressof(blob)
-                #fname = fnames[0]
-                BField = btypes[0]
+        def initialize(blob, init):
+            if is_union:
+                if len(init) > 1:
+                    raise ValueError("union initializer: %d items given, but "
+                                    "only one supported (use a dict if needed)"
+                                     % (len(init),))
+            if not isinstance(init, dict):
+                if isinstance(init, (bytes, unicode)):
+                    raise TypeError("union initializer: got a str")
+                init = tuple(init)
+                if len(init) > len(fnames):
+                    raise ValueError("too many values for %s initializer" %
+                                     CTypesStructOrUnion._get_c_name())
+                init = dict(zip(fnames, init))
+            addr = ctypes.addressof(blob)
+            for fname, value in init.items():
+                BField, bitsize = name2fieldtype[fname]
+                assert bitsize < 0, \
+                       "not implemented: initializer with bit fields"
+                offset = CTypesStructOrUnion._offsetof(fname)
                 PTR = ctypes.POINTER(BField._ctype)
-                BField._initialize(ctypes.cast(addr, PTR).contents, init)
+                p = ctypes.cast(addr + offset, PTR)
+                BField._initialize(p.contents, value)
+        is_union = CTypesStructOrUnion._kind == 'union'
+        name2fieldtype = dict(zip(fnames, zip(btypes, bitfields)))
         #
         for fname, BField, bitsize in fields:
             if hasattr(CTypesStructOrUnion, fname):
@@ -627,6 +754,17 @@ class CTypesBackend(object):
                     return BField._from_ctypes(p.contents)
                 def setter(self, value, fname=fname, BField=BField):
                     setattr(self._blob, fname, BField._to_ctypes(value))
+                #
+                if issubclass(BField, CTypesGenericArray):
+                    setter = None
+                    if BField._declared_length == 0:
+                        def getter(self, fname=fname, BFieldPtr=BField._CTPtr,
+                                   offset=CTypesStructOrUnion._offsetof(fname),
+                                   PTR=ctypes.POINTER(BField._ctype)):
+                            addr = ctypes.addressof(self._blob)
+                            p = ctypes.cast(addr + offset, PTR)
+                            return BFieldPtr._from_ctypes(p)
+                #
             else:
                 def getter(self, fname=fname, BField=BField):
                     return BField._from_ctypes(getattr(self._blob, fname))
@@ -657,27 +795,61 @@ class CTypesBackend(object):
             nameargs.append('...')
         nameargs = ', '.join(nameargs)
         #
-        class CTypesFunction(CTypesGenericPtr):
+        class CTypesFunctionPtr(CTypesGenericPtr):
             __slots__ = ['_own_callback', '_name']
             _ctype = ctypes.CFUNCTYPE(getattr(BResult, '_ctype', None),
                                       *[BArg._ctype for BArg in BArgs],
                                       use_errno=True)
             _reftypename = BResult._get_c_name('(* &)(%s)' % (nameargs,))
 
-            def __init__(self, init):
+            def __init__(self, init, error=None):
                 # create a callback to the Python callable init()
+                import traceback
                 assert not has_varargs, "varargs not supported for callbacks"
+                if getattr(BResult, '_ctype', None) is not None:
+                    error = BResult._from_ctypes(
+                        BResult._create_ctype_obj(error))
+                else:
+                    error = None
                 def callback(*args):
                     args2 = []
                     for arg, BArg in zip(args, BArgs):
                         args2.append(BArg._from_ctypes(arg))
-                    res2 = init(*args2)
-                    return BResult._to_ctypes(res2)
-                self._as_ctype_ptr = CTypesFunction._ctype(callback)
+                    try:
+                        res2 = init(*args2)
+                    except:
+                        traceback.print_exc()
+                        res2 = error
+                    else:
+                        res2 = BResult._to_ctypes(res2)
+                    if issubclass(BResult, CTypesGenericPtr):
+                        if res2:
+                            res2 = ctypes.cast(res2, ctypes.c_void_p).value
+                                # .value: http://bugs.python.org/issue1574593
+                        else:
+                            res2 = None
+                    #print repr(res2)
+                    return res2
+                if issubclass(BResult, CTypesGenericPtr):
+                    # The only pointers callbacks can return are void*s:
+                    # http://bugs.python.org/issue5710
+                    callback_ctype = ctypes.CFUNCTYPE(
+                        ctypes.c_void_p,
+                        *[BArg._ctype for BArg in BArgs],
+                        use_errno=True)
+                else:
+                    callback_ctype = CTypesFunctionPtr._ctype
+                self._as_ctype_ptr = callback_ctype(callback)
                 self._address = ctypes.cast(self._as_ctype_ptr,
                                             ctypes.c_void_p).value
                 self._own_callback = init
 
+            @staticmethod
+            def _initialize(ctypes_ptr, value):
+                if value:
+                    raise NotImplementedError("ctypes backend: not supported: "
+                                          "initializers for function pointers")
+
             def __repr__(self):
                 c_name = getattr(self, '_name', None)
                 if c_name:
@@ -689,8 +861,8 @@ class CTypesBackend(object):
 
             def _get_own_repr(self):
                 if getattr(self, '_own_callback', None) is not None:
-                    return ' calling %r' % (self._own_callback,)
-                return ''
+                    return 'calling %r' % (self._own_callback,)
+                return super(CTypesFunctionPtr, self)._get_own_repr()
 
             def __call__(self, *args):
                 if has_varargs:
@@ -716,12 +888,14 @@ class CTypesBackend(object):
                 result = self._as_ctype_ptr(*ctypes_args)
                 return BResult._from_ctypes(result)
         #
-        CTypesFunction._fix_class()
-        return CTypesFunction
+        CTypesFunctionPtr._fix_class()
+        return CTypesFunctionPtr
 
     def new_enum_type(self, name, enumerators, enumvalues):
+        assert isinstance(name, str)
         mapping = dict(zip(enumerators, enumvalues))
-        reverse_mapping = dict(reversed(zip(enumvalues, enumerators)))
+        reverse_mapping = dict(zip(reversed(enumvalues),
+                                   reversed(enumerators)))
         CTypesInt = self.ffi._get_cached_btype(model.PrimitiveType('int'))
         #
         def forward_map(source):
@@ -742,7 +916,7 @@ class CTypesBackend(object):
             __slots__ = []
             _reftypename = 'enum %s &' % name
 
-            def __str__(self):
+            def _to_string(self, maxlen):
                 return str(CTypesEnum._from_ctypes(self._value))
 
             @classmethod
@@ -772,8 +946,33 @@ class CTypesBackend(object):
     def set_errno(self, value):
         ctypes.set_errno(value)
 
+    def string(self, b, maxlen=-1):
+        return b._to_string(maxlen)
+
     def buffer(self, bptr, size=-1):
+        if sys.version_info >= (3,):
+            # buf = bptr._as_ctype_ptr
+            # return memoryview(buf.contents)
+            if isinstance(bptr, CTypesGenericPtr):
+                buf = bptr._as_ctype_ptr
+                val = buf.contents
+            elif isinstance(bptr, CTypesGenericArray):
+                buf = bptr._blob
+                val = bptr._blob
+            else:
+                raise TypeError(bptr)
+            class Hack(ctypes.Union):
+                _fields_ = [('stupid', type(val))]
+            ptr = ctypes.cast(buf, ctypes.POINTER(Hack))
+            view = memoryview(ptr.contents)
+            if size >= 0:
+                return view.cast('B')[:size]
+            else:
+                return view.cast('B')
+
         # haaaaaaaaaaaack
+        if '__pypy__' in sys.builtin_module_names:
+            raise NotImplementedError("PyPy: ffi.buffer() with ctypes backend")
         call = ctypes.pythonapi.PyBuffer_FromReadWriteMemory
         call.argtypes = (ctypes.c_void_p, ctypes.c_size_t)
         call.restype = ctypes.py_object
@@ -806,13 +1005,15 @@ class CTypesBackend(object):
         return BType._offsetof(fieldname)
 
     def newp(self, BType, source):
-        return BType(source)
+        if not issubclass(BType, CTypesData):
+            raise TypeError
+        return BType._newp(source)
 
     def cast(self, BType, source):
         return BType._cast_from(source)
 
-    def callback(self, BType, source):
-        return BType(source)
+    def callback(self, BType, source, error):
+        return BType(source, error)
 
     typeof = type
 
@@ -833,7 +1034,10 @@ class CTypesLibrary(object):
         return funcobj
 
     def read_variable(self, BType, name):
-        ctypes_obj = BType._ctype.in_dll(self.cdll, name)
+        try:
+            ctypes_obj = BType._ctype.in_dll(self.cdll, name)
+        except AttributeError as e:
+            raise NotImplementedError(e)
         return BType._from_ctypes(ctypes_obj)
 
     def write_variable(self, BType, name, value):
