@@ -7,6 +7,7 @@ _r_define  = re.compile(r"^\s*#\s*define\s+([A-Za-z_][A-Za-z_0-9]*)\s+(.*?)$",
                         re.MULTILINE)
 _r_partial_enum = re.compile(r"\.\.\.\s*\}")
 _r_enum_dotdotdot = re.compile(r"__dotdotdot\d+__$")
+_r_partial_array = re.compile(r"\[\s*\.\.\.\s*\]")
 _parser_cache = None
 
 def _get_parser():
@@ -25,6 +26,8 @@ def _preprocess(csource):
         macroname, macrovalue = match.groups()
         macros[macroname] = macrovalue
     csource = _r_define.sub('', csource)
+    # Replace "[...]" with "[__dotdotdotarray__]"
+    csource = _r_partial_array.sub('[__dotdotdotarray__]', csource)
     # Replace "...}" with "__dotdotdotNUM__}".  This construction should
     # occur only at the end of enums; at the end of structs we have "...;}"
     # and at the end of vararg functions "...);"
@@ -43,13 +46,14 @@ class Parser(object):
         self._declarations = {}
         self._anonymous_counter = 0
         self._structnode2type = weakref.WeakKeyDictionary()
+        self._override = False
 
     def _parse(self, csource):
         # XXX: for more efficiency we would need to poke into the
         # internals of CParser...  the following registers the
         # typedefs, because their presence or absence influences the
         # parsing itself (but what they are typedef'ed to plays no role)
-        csourcelines = []
+        csourcelines = ['typedef int wchar_t;']
         for name in sorted(self._declarations):
             if name.startswith('typedef '):
                 csourcelines.append('typedef int %s;' % (name[8:],))
@@ -60,7 +64,15 @@ class Parser(object):
         ast = _get_parser().parse(csource)
         return ast, macros
 
-    def parse(self, csource):
+    def parse(self, csource, override=False):
+        prev_override = self._override
+        try:
+            self._override = override
+            self._internal_parse(csource)
+        finally:
+            self._override = prev_override
+
+    def _internal_parse(self, csource):
         ast, macros = self._parse(csource)
         # add the macros
         for key, value in macros.items():
@@ -95,8 +107,10 @@ class Parser(object):
     def _parse_decl(self, decl):
         node = decl.type
         if isinstance(node, pycparser.c_ast.FuncDecl):
-            self._declare('function ' + decl.name,
-                          self._get_type(node, name=decl.name))
+            tp = self._get_type(node, name=decl.name)
+            assert isinstance(tp, model.RawFunctionType)
+            tp = self._get_type_pointer(tp)
+            self._declare('function ' + decl.name, tp)
         else:
             if isinstance(node, pycparser.c_ast.Struct):
                 # XXX do we need self._declare in any of those?
@@ -119,29 +133,34 @@ class Parser(object):
                 else:
                     self._declare('variable ' + decl.name, tp)
 
-    def parse_type(self, cdecl, force_pointer=False):
+    def parse_type(self, cdecl, consider_function_as_funcptr=False):
         ast, macros = self._parse('void __dummy(%s);' % cdecl)
         assert not macros
         typenode = ast.ext[-1].type.args.params[0].type
-        return self._get_type(typenode, force_pointer=force_pointer)
+        return self._get_type(typenode,
+                     consider_function_as_funcptr=consider_function_as_funcptr)
 
     def _declare(self, name, obj):
         if name in self._declarations:
             if self._declarations[name] is obj:
                 return
-            raise api.FFIError("multiple declarations of %s" % (name,))
+            if not self._override:
+                raise api.FFIError(
+                    "multiple declarations of %s (for interactive usage, "
+                    "try cdef(xx, override=True))" % (name,))
         assert name != '__dotdotdot__'
         self._declarations[name] = obj
 
     def _get_type_pointer(self, type, const=False):
-        if isinstance(type, model.FunctionType):
-            return type # "pointer-to-function" ~== "function"
+        if isinstance(type, model.RawFunctionType):
+            return type.as_function_pointer()
         if const:
             return model.ConstPointerType(type)
         return model.PointerType(type)
 
     def _get_type(self, typenode, convert_array_to_pointer=False,
-                  force_pointer=False, name=None):
+                  name=None, partial_length_ok=False,
+                  consider_function_as_funcptr=False):
         # first, dereference typedefs, if we have it already parsed, we're good
         if (isinstance(typenode, pycparser.c_ast.TypeDecl) and
             isinstance(typenode.type, pycparser.c_ast.IdentifierType) and
@@ -152,8 +171,9 @@ class Parser(object):
                 if convert_array_to_pointer:
                     return type.item
             else:
-                if force_pointer:
-                    return self._get_type_pointer(type)
+                if (consider_function_as_funcptr and
+                        isinstance(type, model.RawFunctionType)):
+                    return type.as_function_pointer()
             return type
         #
         if isinstance(typenode, pycparser.c_ast.ArrayDecl):
@@ -163,12 +183,10 @@ class Parser(object):
             if typenode.dim is None:
                 length = None
             else:
-                length = self._parse_constant(typenode.dim)
+                length = self._parse_constant(
+                    typenode.dim, partial_length_ok=partial_length_ok)
             return model.ArrayType(self._get_type(typenode.type), length)
         #
-        if force_pointer:
-            return model.PointerType(self._get_type(typenode))
-        #
         if isinstance(typenode, pycparser.c_ast.PtrDecl):
             # pointer type
             const = (isinstance(typenode.type, pycparser.c_ast.TypeDecl)
@@ -209,7 +227,10 @@ class Parser(object):
         #
         if isinstance(typenode, pycparser.c_ast.FuncDecl):
             # a function type
-            return self._parse_function_type(typenode, name)
+            result = self._parse_function_type(typenode, name)
+            if consider_function_as_funcptr:
+                result = result.as_function_pointer()
+            return result
         #
         raise api.FFIError("bad or unsupported type declaration")
 
@@ -229,10 +250,11 @@ class Parser(object):
                 and list(params[0].type.type.names) == ['void']):
             del params[0]
         args = [self._get_type(argdeclnode.type,
-                               convert_array_to_pointer=True)
+                               convert_array_to_pointer=True,
+                               consider_function_as_funcptr=True)
                 for argdeclnode in params]
         result = self._get_type(typenode.type)
-        return model.FunctionType(tuple(args), result, ellipsis)
+        return model.RawFunctionType(tuple(args), result, ellipsis)
 
     def _is_constant_declaration(self, typenode, const=False):
         if isinstance(typenode, pycparser.c_ast.ArrayDecl):
@@ -316,24 +338,32 @@ class Parser(object):
                 # XXX pycparser is inconsistent: 'names' should be a list
                 # of strings, but is sometimes just one string.  Use
                 # str.join() as a way to cope with both.
-                tp.partial = True
-                if not tp.has_c_name():
-                    raise api.CDefError("%s is partial but has no C name"
-                                        % (tp,))
+                self._make_partial(tp)
                 continue
             if decl.bitsize is None:
                 bitsize = -1
             else:
                 bitsize = self._parse_constant(decl.bitsize)
+            self._partial_length = False
+            type = self._get_type(decl.type, partial_length_ok=True)
+            if self._partial_length:
+                self._make_partial(tp)
             fldnames.append(decl.name)
-            fldtypes.append(self._get_type(decl.type))
+            fldtypes.append(type)
             fldbitsize.append(bitsize)
         tp.fldnames = tuple(fldnames)
         tp.fldtypes = tuple(fldtypes)
         tp.fldbitsize = tuple(fldbitsize)
         return tp
 
-    def _parse_constant(self, exprnode):
+    def _make_partial(self, tp):
+        if not isinstance(tp, model.StructType):
+            raise api.CDefError("%s cannot be partial" % (tp,))
+        if not tp.has_c_name():
+            raise api.CDefError("%s is partial but has no C name" % (tp,))
+        tp.partial = True
+
+    def _parse_constant(self, exprnode, partial_length_ok=False):
         # for now, limited to expressions that are an immediate number
         # or negative number
         if isinstance(exprnode, pycparser.c_ast.Constant):
@@ -343,15 +373,37 @@ class Parser(object):
                 exprnode.op == '-'):
             return -self._parse_constant(exprnode.expr)
         #
+        if partial_length_ok:
+            if (isinstance(exprnode, pycparser.c_ast.ID) and
+                    exprnode.name == '__dotdotdotarray__'):
+                self._partial_length = True
+                return None
+        #
         raise api.FFIError("unsupported non-constant or "
                            "not immediately constant expression")
 
     def _get_enum_type(self, type):
+        # See _get_struct_or_union_type() for the reason of the
+        # complicated logic here.  This is still a simplified version,
+        # assuming that it's ok to assume the more complicated cases
+        # don't occur...
+        try:
+            return self._structnode2type[type]
+        except KeyError:
+            pass
         name = type.name
+        if name is None:
+            self._anonymous_counter += 1
+            explicit_name = '$%d' % self._anonymous_counter
+            key = None
+        else:
+            explicit_name = name
+            key = 'enum %s' % (name,)
+            tp = self._declarations.get(key, None)
+            if tp is not None:
+                return tp
+        #
         decls = type.values
-        key = 'enum %s' % (name,)
-        if key in self._declarations:
-            return self._declarations[key]
         if decls is not None:
             enumerators = [enum.name for enum in decls.enumerators]
             partial = False
@@ -367,11 +419,13 @@ class Parser(object):
                 enumvalues.append(nextenumvalue)
                 nextenumvalue += 1
             enumvalues = tuple(enumvalues) 
-            tp = model.EnumType(name, enumerators, enumvalues)
+            tp = model.EnumType(explicit_name, enumerators, enumvalues)
             tp.partial = partial
-            self._declare(key, tp)
+            if key is not None:
+                self._declare(key, tp)
         else:   # opaque enum
             enumerators = ()
             enumvalues = ()
-            tp = model.EnumType(name, (), ())
+            tp = model.EnumType(explicit_name, (), ())
+        self._structnode2type[type] = tp
         return tp
