@@ -13,10 +13,17 @@ class FakeBackend(object):
         return "fake library"
 
     def new_primitive_type(self, name):
-        return FakePrimitiveType(name)
+        return FakeType("primitive " + name)
 
-class FakePrimitiveType(object):
+    def new_void_type(self):
+        return FakeType("void")
+    def new_pointer_type(self, x):
+        return FakeType('ptr-to-%r' % (x,))
+    def cast(self, x, y):
+        return 'casted!'
 
+
+class FakeType(object):
     def __init__(self, cdecl):
         self.cdecl = cdecl
 
@@ -24,5 +31,5 @@ class FakePrimitiveType(object):
 def test_typeof():
     ffi = FFI(backend=FakeBackend())
     clong = ffi.typeof("signed long int")
-    assert isinstance(clong, FakePrimitiveType)
-    assert clong.cdecl == 'long'
+    assert isinstance(clong, FakeType)
+    assert clong.cdecl == 'primitive long'
