@@ -1,12 +1,5 @@
-import sys
-
-import py
-
-if '__pypy__' in sys.modules:
-    py.test.skip("C backend tests are CPython only")
-
 from testing import backend_tests, test_function, test_ownlib
-import _ffi_backend
+import _cffi_backend
 
 
 class TestFFI(backend_tests.BackendTests,
@@ -16,4 +9,4 @@ class TestFFI(backend_tests.BackendTests,
 
     @staticmethod
     def Backend():
-        return _ffi_backend
+        return _cffi_backend
